package dev.fyloz.trial.colorrecipesexplorer.service

import dev.fyloz.trial.colorrecipesexplorer.config.blacklistedJwtTokens
import dev.fyloz.trial.colorrecipesexplorer.config.defaultGroupCookieName
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityAlreadyExistsException
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityNotFoundException
import dev.fyloz.trial.colorrecipesexplorer.model.*
import dev.fyloz.trial.colorrecipesexplorer.model.validation.or
import dev.fyloz.trial.colorrecipesexplorer.repository.EmployeeGroupRepository
import dev.fyloz.trial.colorrecipesexplorer.repository.EmployeeRepository
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.security.core.userdetails.User
import org.springframework.security.core.userdetails.UserDetails
import org.springframework.security.core.userdetails.UserDetailsService
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.transaction.Transactional

interface EmployeeService : ExternalModelService<Employee, EmployeeSaveDto, EmployeeUpdateDto, EmployeeRepository> {
    /** Check if an [Employee] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean

    /** Gets the employee with the given [id]. */
    fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee

    /** Gets all employees which have the given [group]. */
    fun getByGroup(group: EmployeeGroup): Collection<Employee>

    /** Gets the default user of the given [group]. */
    fun getDefaultGroupEmployee(group: EmployeeGroup): Employee

    /** Save a default group employee for the given [group]. */
    fun saveDefaultGroupEmployee(group: EmployeeGroup)

    /** Updates de given [entity]. **/
    fun update(entity: Employee, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee

    /** Updates the last login time of the employee with the given [employeeId]. */
    fun updateLastLoginTime(employeeId: Long, time: LocalDateTime = LocalDateTime.now()): Employee

    /** Updates the password of the employee with the given [id]. */
    fun updatePassword(id: Long, password: String): Employee

    /** Adds the given [permission] to the employee with the given [employeeId]. */
    fun addPermission(employeeId: Long, permission: EmployeePermission): Employee

    /** Removes the given [permission] from the employee with the given [employeeId]. */
    fun removePermission(employeeId: Long, permission: EmployeePermission): Employee

    /** Logout an user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest)
}

interface EmployeeGroupService :
    ExternalNamedModelService<EmployeeGroup, EmployeeGroupSaveDto, EmployeeGroupUpdateDto, EmployeeGroupRepository> {
    /** Gets all the employees of the group with the given [id]. */
    fun getEmployeesForGroup(id: Long): Collection<Employee>

    /** Gets the default group from a cookie in the given HTTP [request]. */
    fun getRequestDefaultGroup(request: HttpServletRequest): EmployeeGroup

    /** Sets the default group cookie for the given HTTP [response]. */
    fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse)

    /** Adds the employee with the given [employeeId] to the group with the given [groupId]. */
    fun addEmployeeToGroup(groupId: Long, employeeId: Long)

    /**
     * Adds a given [employee] to a given [group].
     *
     * If the [employee] is already in the [group], nothing will be done.
     * If the [employee] is already in a group, it will be removed from it.
     */
    fun addEmployeeToGroup(group: EmployeeGroup, employee: Employee)

    /** Removes the employee with the given [employeeId] from the group with the given [groupId]. */
    fun removeEmployeeFromGroup(groupId: Long, employeeId: Long)

    /** Removes a given [employee] from the given [group]. */
    fun removeEmployeeFromGroup(group: EmployeeGroup, employee: Employee)
}

interface EmployeeUserDetailsService : UserDetailsService {
    /** Loads an [User] for the given [employeeId]. */
    fun loadUserByEmployeeId(employeeId: Long, ignoreDefaultGroupUsers: Boolean = false): UserDetails
}

@Service
class EmployeeServiceImpl(employeeRepository: EmployeeRepository, val passwordEncoder: PasswordEncoder) :
    AbstractExternalModelService<Employee, EmployeeSaveDto, EmployeeUpdateDto, EmployeeRepository>(employeeRepository),
    EmployeeService {
    @Autowired
    lateinit var groupService: EmployeeGroupServiceImpl

    override fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean =
        repository.existsByFirstNameAndLastName(firstName, lastName)

    override fun getAll(): Collection<Employee> =
        super.getAll().filter { !it.isSystemUser && !it.isDefaultGroupUser }

    override fun getById(id: Long): Employee =
        getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee =
        super.getById(id).apply {
            if (ignoreSystemUsers && isSystemUser || ignoreDefaultGroupUsers && isDefaultGroupUser) throw EntityNotFoundException(
                id
            )
        }

    override fun getByGroup(group: EmployeeGroup): Collection<Employee> =
        repository.findAllByGroup(group).filter {
            !it.isSystemUser && !it.isDefaultGroupUser
        }

    override fun getDefaultGroupEmployee(group: EmployeeGroup): Employee =
        repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)

    override fun save(entity: EmployeeSaveDto): Employee =
        save(with(entity) {
            Employee(
                id,
                firstName,
                lastName,
                passwordEncoder.encode(password),
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = if (groupId != null) groupService.getById(groupId!!) else null,
                permissions = permissions
            )
        })

    override fun save(entity: Employee): Employee {
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw EntityAlreadyExistsException("${entity.firstName} ${entity.lastName}")
        return super<AbstractExternalModelService>.save(entity)
    }

    override fun saveDefaultGroupEmployee(group: EmployeeGroup) {
        save(
            Employee(
                id = 1000000L + group.id!!,
                firstName = group.name,
                lastName = "EmployeeModel",
                password = passwordEncoder.encode(group.name),
                group = group,
                isDefaultGroupUser = true
            )
        )
    }

    override fun updateLastLoginTime(employeeId: Long, time: LocalDateTime): Employee {
        val employee = getById(employeeId, ignoreDefaultGroupUsers = true, ignoreSystemUsers = false)
        employee.lastLoginTime = time
        return update(
            employee,
            ignoreDefaultGroupUsers = true,
            ignoreSystemUsers = false
        )
    }

    override fun update(entity: EmployeeUpdateDto): Employee {
        val persistedEmployee by lazy { getById(entity.id) }
        return update(with(entity) {
            Employee(
                id = id,
                firstName = firstName or persistedEmployee.firstName,
                lastName = lastName or persistedEmployee.lastName,
                password = persistedEmployee.password,
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = persistedEmployee.group,
                permissions = permissions?.toMutableSet() ?: persistedEmployee.permissions,
                lastLoginTime = persistedEmployee.lastLoginTime
            )
        })
    }

    override fun update(entity: Employee): Employee =
        update(entity, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun update(entity: Employee, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee {
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw EntityAlreadyExistsException("${entity.firstName} ${entity.lastName}")
        }

        return super<AbstractExternalModelService>.update(entity)
    }

    override fun updatePassword(id: Long, password: String): Employee {
        val persistedEmployee = getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
        return super<AbstractExternalModelService>.update(with(persistedEmployee) {
            Employee(
                id,
                firstName,
                lastName,
                passwordEncoder.encode(password),
                isDefaultGroupUser,
                isSystemUser,
                group,
                permissions,
                lastLoginTime
            )
        })
    }

    override fun addPermission(employeeId: Long, permission: EmployeePermission): Employee =
        super<AbstractExternalModelService>.update(getById(employeeId).apply { permissions += permission })

    override fun removePermission(employeeId: Long, permission: EmployeePermission): Employee =
        super<AbstractExternalModelService>.update(getById(employeeId).apply { permissions -= permission })

    override fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, "Authorization")
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }
}

const val defaultGroupCookieMaxAge = 10 * 365 * 24 * 60 * 60 // 10 ans

@Service
class EmployeeGroupServiceImpl(
    val employeeService: EmployeeService,
    employeeGroupRepository: EmployeeGroupRepository
) :
    AbstractExternalNamedModelService<EmployeeGroup, EmployeeGroupSaveDto, EmployeeGroupUpdateDto, EmployeeGroupRepository>(
        employeeGroupRepository
    ),
    EmployeeGroupService {
    override fun existsByName(name: String): Boolean = repository.existsByName(name)
    override fun getEmployeesForGroup(id: Long): Collection<Employee> =
        getById(id).employees

    @Transactional
    override fun save(entity: EmployeeGroup): EmployeeGroup {
        return super<AbstractExternalNamedModelService>.save(entity).apply {
            employeeService.saveDefaultGroupEmployee(this)
        }
    }

    override fun update(entity: EmployeeGroupUpdateDto): EmployeeGroup {
        val persistedGroup by lazy { getById(entity.id) }
        return update(with(entity) {
            EmployeeGroup(
                entity.id,
                if (name.isNotBlank()) entity.name else persistedGroup.name,
                if (permissions.isNotEmpty()) entity.permissions else persistedGroup.permissions,
                persistedGroup.employees
            )
        })
    }

    @Transactional
    override fun delete(entity: EmployeeGroup) {
        employeeService.delete(employeeService.getDefaultGroupEmployee(entity))
        super.delete(entity)
    }

    override fun getRequestDefaultGroup(request: HttpServletRequest): EmployeeGroup {
        val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
            ?: throw EntityNotFoundException("defaultGroup")
        val defaultGroupUser = employeeService.getById(
            defaultGroupCookie.value.toLong(),
            ignoreDefaultGroupUsers = false,
            ignoreSystemUsers = true
        )
        return defaultGroupUser.group!!
    }

    override fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse) {
        val group = getById(groupId)
        val defaultGroupUser = employeeService.getDefaultGroupEmployee(group)
        response.addHeader(
            "Set-Cookie",
            "$defaultGroupCookieName=${defaultGroupUser.id}; Max-Age=${defaultGroupCookieMaxAge}; Path=/api; HttpOnly; Secure; SameSite=strict"
        )
    }

    override fun addEmployeeToGroup(groupId: Long, employeeId: Long) {
        addEmployeeToGroup(getById(groupId), employeeService.getById(employeeId))
    }

    @Transactional
    override fun addEmployeeToGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group == group) return
        if (employee.group != null) removeEmployeeFromGroup(employee.group!!, employee)

        group.employees.add(employee)
        employee.group = group
        update(group)
        employeeService.update(employee)
    }

    override fun removeEmployeeFromGroup(groupId: Long, employeeId: Long) =
        removeEmployeeFromGroup(getById(groupId), employeeService.getById(employeeId))

    @Transactional
    override fun removeEmployeeFromGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group == null || employee.group != group) return

        group.employees.removeIf { it.id == employee.id }
        employee.group = null
        update(group)
        employeeService.update(employee)
    }
}

@Service
class EmployeeUserDetailsServiceImpl(
    val employeeService: EmployeeService
) :
    EmployeeUserDetailsService {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserByEmployeeId(username.toLong(), true)
        } catch (ex: EntityNotFoundException) {
            throw UsernameNotFoundException(username)
        } catch (ex: EntityNotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    override fun loadUserByEmployeeId(employeeId: Long, ignoreDefaultGroupUsers: Boolean): UserDetails {
        val employee = employeeService.getById(
            employeeId,
            ignoreDefaultGroupUsers = ignoreDefaultGroupUsers,
            ignoreSystemUsers = false
        )
        return User(employee.id.toString(), employee.password, employee.getAuthorities())
    }
}
