package dev.fyloz.trial.colorrecipesexplorer.service

import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantitiesException
import dev.fyloz.trial.colorrecipesexplorer.exception.LowQuantityException
import dev.fyloz.trial.colorrecipesexplorer.model.MaterialQuantityDto
import dev.fyloz.trial.colorrecipesexplorer.service.utils.filterThrows
import org.springframework.stereotype.Service
import javax.transaction.Transactional

interface InventoryService {
    /** Adds each given [MaterialQuantityDto] to the inventory. */
    fun add(materialQuantities: Collection<MaterialQuantityDto>)

    /** Adds a given quantity to the given [Material]'s inventory quantity according to the given [materialQuantity]. */
    fun add(materialQuantity: MaterialQuantityDto)

    /** Deducts the inventory quantity of each given [MaterialQuantityDto]. */
    fun deduct(materialQuantities: Collection<MaterialQuantityDto>)

    /** Deducts the inventory quantity of a given [Material] by a given quantity according to the given [materialQuantity]. */
    fun deduct(materialQuantity: MaterialQuantityDto)
}

@Service
class InventoryServiceImpl(
    private val materialService: MaterialService
) : InventoryService {
    @Transactional
    override fun add(materialQuantities: Collection<MaterialQuantityDto>) {
        materialQuantities.forEach(::add)
    }

    override fun add(materialQuantity: MaterialQuantityDto) {
        materialService.updateQuantity(
            materialService.getById(materialQuantity.material),
            materialQuantity.quantity
        )
    }

    @Transactional
    override fun deduct(materialQuantities: Collection<MaterialQuantityDto>) {
        with(materialQuantities.filterThrows<MaterialQuantityDto, LowQuantityException> {
            deduct(it)
        }) {
            if (this.isNotEmpty()) {
                throw LowQuantitiesException(this)
            }
        }
    }

    override fun deduct(materialQuantity: MaterialQuantityDto) {
        val material = materialService.getById(materialQuantity.material)
        if (material.inventoryQuantity >= materialQuantity.quantity) {
            materialService.updateQuantity(
                material,
                -materialQuantity.quantity
            )
        } else {
            throw LowQuantityException(materialQuantity)
        }
    }
}
